#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "psg.h"
#include "internal.h"

/* Period counter for seamless frames */
struct frame tdata;

/* Values stolen from STonX */
static int amps[16]={0,1,1,2,3,4,5,7,12,20,28,44,70,110,165,255};

static int16_t _clip(int input)
{
  if(input < -32768) return (int16_t)-32768;
  if(input > 32767) return (int16_t) 32767;
  return input;
}

psg *psg_init(int mode, int bits, int freq)
{
  psg *psg_struct;
  int i;

  if((bits != 16)) {
    fprintf(stderr, "Unsupported bitvalue\n");
    return (psg *)NULL;
  }

  if((mode < 0) || (mode > 1)) {
    fprintf(stderr, "Unsupported mode\n");
    return (psg *)NULL;
  }

  if(freq != 44100 && freq != 22050) {
    fprintf(stderr, "Unsupported frequency\n");
    return (psg *)NULL;
  }

  psg_struct = (psg *) malloc(sizeof(psg));

  psg_struct->output.mode = mode;
  psg_struct->output.bits = bits;
  psg_struct->output.freq = freq;
  tdata.sample_freq = freq;

  for(i=0;i<3;i++) {
    tdata.wave_period[i] = 0;
  }
  
  return psg_struct;
}

static
int16_t _get_one_tonesample(int offset)
{
  if((offset&(QUALITY-1)) < (QUALITY/2))
    return SAMPLE_MAX;
  else 
    return SAMPLE_MIN;
}

static
int16_t _get_one_noisesample(int offset)
{
  tdata.noisecnt--;
  if(tdata.noisecnt < 0) {
    tdata.noiseval = (NOISE_MAX&rand())-SAMPLE_MAX;
    tdata.noisecnt = tdata.noiselev;
  }
  return tdata.noiseval;
}

static
int _get_real_freq(int psgfreq, int sample_freq)
{
  int denominator;

  if(psgfreq)
    {
      denominator = sample_freq * psgfreq;
      return (QUALITY *(PSG_FREQUENCY/16.0) + (denominator >> 1)) / denominator;
    }
  else
    return 0;
}

void _get_values(psg *psg)
{
  int i, tmp;
  
  for(i=0;i<3;i++) {
    if(psg->regs[8+i] == 16) {
      tdata.rvol[i] = 0;
      tdata.env[i] = 1;
      if(psg->regs[13] != 0xff) {
	tdata.env_period = 0;
	tdata.env_mode = psg->regs[13];
      }
    } else {
      tdata.rvol[i] = psg->regs[8+i];
      tdata.env[i] = 0;
    }
    tmp = psg->regs[11] + (psg->regs[12] << 8);
    if(tmp)
      tdata.env_freq = (double) (PSG_FREQUENCY/(256 * tmp));
    else
      tdata.env_freq = 0;
    tdata.rfreq[i] = _get_real_freq((psg->regs[1+i*2]<<8) + psg->regs[0+i*2], psg->output.freq);
    tdata.tone[i] = !(psg->regs[7] & (1<<i));
    tdata.noise[i] = !(psg->regs[7] & (8<<i));
    tdata.noiselev = psg->regs[6] & 0x1f;
  }
}

static
int16_t _get_one_sample(void)
{
  int i, count, tvol;
  int32_t tmp[3];
  int32_t tmp2;

  tmp2 = 0;
  tvol = 0;
  count = 0;

  for(i=0;i<3;i++) {
    tmp[i] = 0;
    if(tdata.tone[i] && !tdata.noise[i]) {
      if(tdata.env[i]) {
	tmp[i] = _clip(tmp[i] + ((_get_one_tonesample(tdata.wave_period[i])*
			   amps[(int)
			       (15*env_wave(tdata.env_period*M_PI*2/QUALITY, 
					    tdata.env_mode))]>>8)));
      } else {
	tmp[i] = _clip(tmp[i] + ((_get_one_tonesample(tdata.wave_period[i])*
			    amps[tdata.rvol[i]])>>8));
      }
      tdata.wave_period[i] += tdata.rfreq[i];
    }
    if(tdata.noise[i] && !tdata.tone[i]) {
      if(tdata.env[i]) {
	tmp[i] = _clip(tmp[i] + ((_get_one_noisesample(tdata.wave_period[i])*
			   amps[(int)
			       (15*env_wave(tdata.env_period*M_PI*2/QUALITY, 
					    tdata.env_mode))]>>8)));
      } else {
	tmp[i] = _clip(tmp[i] + ((_get_one_noisesample(tdata.wave_period[i])*
			    amps[tdata.rvol[i]])>>8));
      }
      //      tdata.wave_period[i] += tdata.rfreq[i];
    }
    if(tdata.noise[i] && tdata.tone[i]) {
      if(tdata.env[i]) {
	tmp[i] = _clip(tmp[i] + ((_clip(_get_one_tonesample(tdata.wave_period[i])+
				 _get_one_noisesample(tdata.wave_period[i]))*
			    amps[(int)
				(15*env_wave(tdata.env_period*M_PI*2/QUALITY, 
					     tdata.env_mode))]>>8)));
      } else {
	tmp[i] = _clip(tmp[i] + (_clip(_get_one_tonesample(tdata.wave_period[i])+
				 _get_one_noisesample(tdata.wave_period[i]))*
			   amps[tdata.rvol[i]]>>8));
      }
      tdata.wave_period[i] += tdata.rfreq[i];
    }
    if(!tdata.noise[i] && !tdata.tone[i]) {
      if(tdata.env[i]) {
	tmp[i] = _clip(tmp[i] + ((SAMPLE_MAX*
			   amps[(int)
			       (15*env_wave(tdata.env_period*M_PI*2/QUALITY, 
					    tdata.env_mode))]>>8)));
      } else {
	tmp[i] = _clip(tmp[i] + ((SAMPLE_MAX*
			    amps[tdata.rvol[i]])>>8));
      }
    }
  }
  if(tdata.env_freq)
    tdata.env_period += 2*(QUALITY/(tdata.sample_freq * tdata.env_freq));

  tmp2 = _clip(((double)((tmp[0]+tmp[1]+tmp[2]))*0.3f));
  
  return tmp2;
}

static
int16_t old_get_one_sample(void)
{
  int i, count, tvol;
  int32_t tmp;
  double tmp2;

  tmp = 0;
  tmp2 = 0;
  tvol = 0;
  count = 0;

  for(i=0;i<3;i++) {
    if(tdata.tone[i]) {
      tmp += (_get_one_tonesample(tdata.wave_period[i])*
	      amps[tdata.rvol[i]])>>8;
      tdata.wave_period[i] += tdata.rfreq[i];
      count++;
    }
    if(tdata.noise[i]) {
      tmp += (_get_one_noisesample(tdata.wave_period[i])*
	      amps[tdata.rvol[i]]);
      tdata.wave_period[i] += tdata.rfreq[i];
      count++;
    }
    if(tdata.env[i]) {
      tmp *= env_wave(tdata.env_period*M_PI*2/QUALITY, tdata.env_mode);
    }
  }
  if(tdata.env_freq)
    tdata.env_period += 2*(QUALITY/(tdata.sample_freq * tdata.env_freq));

  return _clip(tmp);
}

int psg_create_samples(void *buffer, psg *psg, int usec)
{
  int16_t *tmp;
  int samples, i, tval;

  tmp = (int16_t *) buffer;
  samples = psg->output.freq*usec/1000000;

  _get_values(psg);
  
  for(i=0;i<samples;i++) {
    tval = _get_one_sample();
    *tmp++ = tval;
    if(psg->output.mode)
      *tmp++ = tval;
  }
  
  return 0;
}
